/*****************************************************************************\
*                                                                             *
* callout.c   -- Interface for the "Callout" control                          *
*                                                                             *
* By Thiadmer Riemersma, for Windows Developer Magazine                       *
*                                                                             *
\*****************************************************************************/
#define STRICT
#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>

#if defined CALLOUT_DLL
  #define COMPILE_MULTIMON_STUBS
#endif
#include <multimon.h>

#include "callout.h"

typedef int (__stdcall *DRAWTEXT)(HDC, LPCTSTR, int, LPRECT, UINT);

#if defined HTMLSUPPORT
  int __stdcall DrawHTML(HDC hdc, LPCTSTR lpString, int nCount, LPRECT lpRect, UINT uFormat);
#endif

struct tagCalloutInfo {
  HFONT    hfont;
  HICON    hicon;
  POINT    ptAnchor;    /* coordinates of the anchor (screen coordinates) */
  int      minwidth;
  int      extraheight;
  int      radius;
  int      border;
  int      tailstyle;
  int      tailaspect;
  int      tailheight;
  int      tailoffset;
  int      tailjoin;
  int      joinwidth;
  BOOL     alignbelow;
  COLORREF BackColor;
  COLORREF TextColor;
  DWORD    timeout;
  DRAWTEXT DrawText;
  /* calculated properties */
  POINT    ptBalloon;   /* coordinates of the top-left corner of the balloon */
  RECT     rcText;      /* rectangle of the text relative to the window */
  RECT     rcBalloon;   /* rectangle of the balloon relative to the window */
  int      joinpos;     /* horizontal pos. where the tail joins the balloon */
  int      extent;      /* horizontal tail extent */
  DWORD    timestamp;   /* time that the callout became visible */
  /* resources to delete */
  HBRUSH   hbrBackgr;
  HFONT    hfontOrg;
};
typedef struct tagCalloutInfo CalloutInfo;

#define ATOM_NAME       "Callout:info"
#define MARGIN_HOR      4
#define MARGIN_TOP      2       /* the internal leading in the font is on top */
#define MARGIN_BOTTOM   3
#define BALLOON_SPACING 2       /* spacing between two balloons */
#define TIMER_ID        1

#if !defined ICON_SMALL
  #define ICON_SMALL    0
  #define ICON_BIG      1
#endif

#if !defined UNUSED_PARAM
  #define UNUSED_PARAM(p) ((void)(p))
#endif


#define GETFIELD(f)     (((ci=(CalloutInfo*)GetProp(hwnd,ATOM_NAME))!=NULL) \
                          ? ci->f : 0)
#define SETFIELD(f,p,r) if ((ci=(CalloutInfo*)GetProp(hwnd,ATOM_NAME))!=NULL) \
                        { ci->f = (p);                       \
                          CalcShowCallout(hwnd, (r), FALSE); \
                        }
#define HANDLEPROP(m,f,p,r) \
                        case CM_GET ## m : return (LRESULT)GETFIELD(f); \
                        case CM_SET ## m : SETFIELD(f,p,r); return 0


static
HWND GetChildParent(HWND hwnd)
{
  /* check the style: call GetParent() only for child windows because
   * GetParent() returns the "owner" window for popup windows
   */
  DWORD Style = GetWindowLong(hwnd, GWL_STYLE);
  return ((Style & WS_CHILD) == WS_CHILD) ? GetParent(hwnd) : NULL;
}

/* Get the dimension of the text field, optionally reformatting it so that we
 * get a nice aspect ratio
 */
static
void CalcTextRect(CalloutInfo *ci, HWND hwnd, LPCSTR text)
{
  HDC hdc;
  HFONT hfontOrg;

  assert(ci != NULL);
  SetRect(&ci->rcText, 0, 0, 0, 0);

  /* when passed a NULL or empty string, we are done quickly */
  if (text != NULL && strlen(text) > 0) {

    /* set up the Display Context */
    hdc = GetDC(hwnd);
    hfontOrg = SelectObject(hdc, ci->hfont);

    /* first calculate the width/height of the text without word wrapping */
    ci->DrawText(hdc, text, -1, &ci->rcText, DT_CALCRECT | DT_NOPREFIX);

    /* as we want a maximal width:height aspect ratio of 4:1, we adjust the
     * width of the output rectangle otherwise.
     */
    if (ci->rcText.right > 4 * ci->rcText.bottom) {
      long area = ci->rcText.right * ci->rcText.bottom;
      ci->rcText.right = 2 * (int)sqrt(area);
      /* but the width is clamped to some minimum */
      if (ci->rcText.right <= ci->minwidth)
        ci->rcText.right = ci->minwidth;
    } /* if */
    /* reformat the text to fit in the calculated width */
    ci->DrawText(hdc, text, -1, &ci->rcText,
                 DT_CALCRECT | DT_NOPREFIX | DT_WORDBREAK);

    /* no longer need the Display Context */
    SelectObject(hdc, hfontOrg);
    ReleaseDC(hwnd, hdc);

  } /* if */

  /* The call to DrawText() can return a narrower margin than the minimum
   * width, so check and adjust the minimum width again
   */
  if (ci->rcText.right <= ci->minwidth)
    ci->rcText.right = ci->minwidth;
  ci->rcText.bottom += ci->extraheight;

  assert(ci->rcText.left == 0 && ci->rcText.top == 0);
}

/* Calculate the bonding rectangle of the balloon, and shift the text rectangle
 * to fit at its correct position inside the balloon.
 */
static
void CalcBalloonRect(CalloutInfo *ci)
{
  int iconspace;

  assert(ci != NULL);
  iconspace = ci->hicon ? GetSystemMetrics(SM_CXSMICON) : 0;

  OffsetRect(&ci->rcText,
             ci->radius + ci->border + MARGIN_HOR + iconspace,
             ci->radius + ci->border + MARGIN_TOP);
  SetRect(&ci->rcBalloon,
          ci->rcText.left   - ci->radius - ci->border - MARGIN_HOR - iconspace,
          ci->rcText.top    - ci->radius - ci->border - MARGIN_TOP,
          ci->rcText.right  + ci->radius + ci->border + MARGIN_HOR,
          ci->rcText.bottom + ci->radius + ci->border + MARGIN_BOTTOM);
}

static
void ClampMonitor(HMONITOR hmon, int *x, int *y, int width, int height)
{
  MONITORINFO mi;
  mi.cbSize =  sizeof mi;
  GetMonitorInfo(hmon, &mi);
  if (*x < mi.rcWork.left)
    *x = mi.rcWork.left;
  else if (*x + width > mi.rcWork.right)
    *x = mi.rcWork.right - width;
  if (*y < mi.rcWork.top)
    *y = mi.rcWork.top;
  else if (*y + height > mi.rcWork.bottom)
    *y = mi.rcWork.bottom - height;
}

static
void ClampParent(HWND hwnd, int *x, int *y, int width, int height)
{
  /* check the style: call GetParent() only for child windows because
   * GetParent() returns the "owner" window for popup windows
   */
  HWND hwndParent = GetChildParent(hwnd);
  if (hwndParent) {
    RECT rcParent;
    GetClientRect(hwndParent, &rcParent);
    ClientToScreen(hwndParent, (LPPOINT)&rcParent.left);
    ClientToScreen(hwndParent, (LPPOINT)&rcParent.right);
    if (*x < rcParent.left)
      *x = rcParent.left;
    else if (*x + width > rcParent.right)
      *x = rcParent.right - width;
    if (*y < rcParent.top)
      *y = rcParent.top;
    else if (*y + height > rcParent.bottom)
      *y = rcParent.bottom - height;
  } /* if */
}

static
BOOL CheckOverlap(HWND hwnd, int x, int y, int width, int height, int *delta)
{
  RECT rcSibling;
  HWND hwndParent = GetChildParent(hwnd);
  HWND hwndSibling = FindWindowEx(hwndParent, NULL, CALLOUT_CLASS, NULL);
  while (hwndSibling != NULL) {
    if (hwndSibling != hwnd && IsWindowVisible(hwndSibling)) {
      GetWindowRect(hwndSibling, &rcSibling);
      if (x < rcSibling.right && rcSibling.left < x + width
          && y < rcSibling.bottom && rcSibling.top < y + height)
      {
        /* the balloon overlaps another balloon, check whether the sibling
         * is (mostly) on the left or (mostly) on the right
         */
        if (delta) {
          int left = rcSibling.right - x;
          int right = rcSibling.left - (x + width);
          *delta = ((abs(left) <= abs(right)) ? left : right) + BALLOON_SPACING;
        } /* if */
        return TRUE;
      } /* if */
    } /* if */
    hwndSibling = FindWindowEx(hwndParent, hwndSibling, CALLOUT_CLASS, NULL);
  } /* while */

  return FALSE;
}

static
void FindBalloonPosition(CalloutInfo *ci, HWND hwnd)
{
  /* in case this is a multi-monitor system, get the monitor that the anchor
   * point is in */
  HMONITOR hmon = MonitorFromPoint(ci->ptAnchor, MONITOR_DEFAULTTONEAREST);
  int alignabove, offsx, offsy;
  int tailwidth, taildepth, joinmargin;
  int left = 0, top = 0;
  int run;

  assert(ci != NULL);
  tailwidth = ((ci->tailheight + ci->tailoffset) * ci->tailaspect) / 100;
  taildepth = ci->tailheight + ci->tailoffset;
  alignabove = ci->alignbelow;  /* this field gets toggled inside the loop */
  for (run = 0; run < 2; run++) {
    int delta;

    assert(ci->rcBalloon.left == 0 && ci->rcBalloon.top == 0);

    /* create the preferred alignment */
    left = ci->ptAnchor.x
           - (ci->tailjoin * ci->rcBalloon.right) / 100
           - tailwidth;

    /* alignment on top/bottom */
    alignabove = !alignabove;       /* next iteration, try other alignment */
    if (alignabove)
      top = ci->ptAnchor.y - ci->rcBalloon.bottom - taildepth;
    else
      top = ci->ptAnchor.y + taildepth;

    /* clamp inside the monitor and inside any parent window, then verify
     * other constraints
     */
    ClampMonitor(hmon, &left, &top, ci->rcBalloon.right, ci->rcBalloon.bottom);
    ClampParent(hwnd, &left, &top, ci->rcBalloon.right, ci->rcBalloon.bottom);
    if (alignabove && top + ci->rcBalloon.bottom > ci->ptAnchor.y - taildepth) {
      /* cannot align "on top" at all, skip other tests and align immediately
       * on the bottom
       */
      continue;
    } else if (!alignabove && top < ci->ptAnchor.y + taildepth) {
      /* cannot align "on bottom" at all, skip other tests and align
       * immediately on the top
       */
      continue;
    } else if (CheckOverlap(hwnd, left, top,
                            ci->rcBalloon.right, ci->rcBalloon.bottom, &delta))
    { /* apply the delta, then clamp inside the monitor/parent window and
       * check for overlaps again
       */
      left += delta;
      ClampMonitor(hmon, &left, &top, ci->rcBalloon.right,ci->rcBalloon.bottom);
      ClampParent(hwnd, &left, &top, ci->rcBalloon.right, ci->rcBalloon.bottom);
      if (CheckOverlap(hwnd, left, top,
                       ci->rcBalloon.right, ci->rcBalloon.bottom, NULL))
        continue; /* try aligning after toggling top/bottom */
    } /* if */

    /* if we arrive here, the balloon fits in the screen and does not overlap
     * another balloon --meaning that we are done
     */
    break;
  } /* for */

  /* if we arrive here because all alignments that we tried have failed, give
   * up and use the latest alignment that we tried
   */
  ci->ptBalloon.x = left;
  ci->ptBalloon.y = top;

  /* get the attachment point of the tail (relative to the balloon) */
  joinmargin = ci->border + ci->radius + ci->joinwidth;
  ci->joinpos = ci->ptAnchor.x - tailwidth - left;
  if (ci->joinpos < joinmargin)
    ci->joinpos = joinmargin;
  else if (ci->joinpos > ci->rcBalloon.right - joinmargin)
    ci->joinpos = ci->rcBalloon.right - joinmargin;

  /* with the clamped attachment point and the original anchor point, we can
   * now calculate the final width of the tail
   */
  if (taildepth > 0)
    ci->extent = (ci->tailheight * (ci->ptAnchor.x - (left + ci->joinpos)))
                 / taildepth;
  else
    ci->extent = 0;

  /* check whether to move the balloon box and the text box */
  offsx = ((ci->joinpos + ci->extent) < 0)
          ? -(ci->joinpos + ci->extent) : 0;
  offsy = alignabove ? 0 : taildepth;
  OffsetRect(&ci->rcBalloon, offsx, offsy);
  OffsetRect(&ci->rcText, offsx, offsy);
}

static HRGN MakeTailEllipse(LPPOINT Tail, float position)
{
  float centre_y = position * Tail[1].y + (1.0f - position) * Tail[0].y;
  float aspect = fabs((Tail[1].y - Tail[0].y) / (6. * (Tail[3].x - Tail[0].x)));
  RECT rc;
  rc.left   = (int)ceil(position * Tail[1].x + (1.0f - position) * Tail[0].x);
  rc.right  = (int)floor(position * Tail[1].x + (1.0f - position) * Tail[3].x);
  rc.top    = (int)(centre_y - aspect * (rc.right - rc.left));
  rc.bottom = (int)(centre_y + aspect * (rc.right - rc.left));
  return CreateEllipticRgnIndirect(&rc);
}

static HRGN MakeBalloonRegion(CalloutInfo *ci)
{
  HRGN hrgnBalloon;

  /* create the region for the balloon */
  if (ci->radius <= 0)
    hrgnBalloon = CreateRectRgn(ci->rcBalloon.left, ci->rcBalloon.top,
                                ci->rcBalloon.right + 1,
                                ci->rcBalloon.bottom + 1);
  else
    hrgnBalloon = CreateRoundRectRgn(ci->rcBalloon.left, ci->rcBalloon.top,
                                     ci->rcBalloon.right + 1,
                                     ci->rcBalloon.bottom + 1,
                                     2 * ci->radius, 2 * ci->radius);

  if (ci->tailheight > 0) {
    POINT Tail[4];
    HRGN hrgnTail;

    /* create the general triangular shape of the tail */
    Tail[1].x = ci->rcBalloon.left + ci->joinpos + ci->extent;
    if (ci->rcBalloon.top > 0) {
      Tail[1].y = Tail[2].y = 0;
      Tail[0].y = Tail[3].y = ci->rcBalloon.top;
    } else {
      Tail[1].y = Tail[2].y = ci->rcBalloon.bottom + ci->tailheight;
      Tail[0].y = Tail[3].y = ci->rcBalloon.bottom;
    } /* if */
    Tail[0].x = ci->rcBalloon.left + ci->joinpos - ci->joinwidth;
    Tail[2].x = Tail[1].x + 1;
    Tail[3].x = ci->rcBalloon.left + ci->joinpos + ci->joinwidth + 1;
    /* adjust for very narrow balloons */
    if (Tail[0].x < ci->rcBalloon.left + ci->radius)
      Tail[0].x = ci->rcBalloon.left + ci->radius;
    if (Tail[3].x > ci->rcBalloon.right - ci->radius)
      Tail[3].x = ci->rcBalloon.right - ci->radius;

    /* create the region for the tail */
    switch (ci->tailstyle) {
    case CS_SPEAK:
      hrgnTail = CreatePolygonRgn(Tail, sizeof Tail/sizeof Tail[0], ALTERNATE);
      /* create combined region and delete the temporary object */
      CombineRgn(hrgnBalloon, hrgnBalloon, hrgnTail, RGN_OR);
      DeleteObject(hrgnTail);
      break;

    case CS_THINK: {
      /* make three ellipses with centres at 1/10, 4/10 and 7/10 tail height */
      hrgnTail = MakeTailEllipse(Tail, 0.09f);
      CombineRgn(hrgnBalloon, hrgnBalloon, hrgnTail, RGN_OR);
      DeleteObject(hrgnTail);
      hrgnTail = MakeTailEllipse(Tail, 0.4f);
      CombineRgn(hrgnBalloon, hrgnBalloon, hrgnTail, RGN_OR);
      DeleteObject(hrgnTail);
      hrgnTail = MakeTailEllipse(Tail, 0.6f);
      CombineRgn(hrgnBalloon, hrgnBalloon, hrgnTail, RGN_OR);
      DeleteObject(hrgnTail);
      break;
    } /* case */
    } /* switch */

  } /* if */

  return hrgnBalloon;
}

static void CalcShowCallout(HWND hwnd, BOOL Repaint, BOOL ForceRecalc)
{
  CalloutInfo *ci = (CalloutInfo*)GetProp(hwnd, ATOM_NAME);
  LPSTR text;
  int textlength;
  int width, height;
  POINT origin;
  HRGN hrgn;
  HWND hwndParent;

  /* step 1: calculate the bounding box of the text */
  textlength = GetWindowTextLength(hwnd);
  if (textlength > 0) {
    text = malloc(textlength + 1);
    if (text)
      GetWindowText(hwnd, text, textlength + 1);
  } else {
    text = NULL;
  } /* if */
  CalcTextRect(ci, hwnd, text);
  if (text)
    free(text);

  /* step 2: get the bounding box of the balloon around the text */
  CalcBalloonRect(ci);

  /* if the callout window is hidden, do nothing more */
  if (!IsWindowVisible(hwnd) && !ForceRecalc)
    return;

  /* step 3: position the balloon in an intelligent way */
  FindBalloonPosition(ci, hwnd);

  /* step 4: position the window and set its size */
  origin.x = ci->ptBalloon.x + min(ci->joinpos + ci->extent, 0);
  origin.y = ci->ptBalloon.y - ci->rcBalloon.top;
  if ((hwndParent=GetChildParent(hwnd)) != NULL)
    ScreenToClient(hwndParent, &origin);
  width = max(ci->joinpos + ci->extent, ci->rcBalloon.right);
  height = (ci->rcBalloon.bottom - ci->rcBalloon.top) + ci->tailheight;
  SetWindowPos(hwnd, NULL, origin.x, origin.y, width, height,
               SWP_NOZORDER | SWP_NOREDRAW | SWP_NOACTIVATE);

  /* step 5: create and set the region for the callout */
  hrgn = MakeBalloonRegion(ci);
  assert(hrgn!=NULL);
  SetWindowRgn(hwnd, hrgn, Repaint);
}


static BOOL Cls_OnCreate(HWND hwnd, CREATESTRUCT FAR* lpCreateStruct)
{
  CalloutInfo *ci;
  DWORD Style;

  /* check/reset the style of the window */
  Style = GetWindowLong(hwnd, GWL_STYLE);
  if ((Style & (WS_BORDER | WS_DLGFRAME | WS_THICKFRAME)) != 0)
    SetWindowLong(hwnd, GWL_STYLE,
                  Style & ~(WS_BORDER | WS_DLGFRAME | WS_THICKFRAME));

  ci = malloc(sizeof *ci);
  if (ci) {
    HDC hdc = GetDC(hwnd);
    memset(ci, 0, sizeof *ci);
    ci->hfont = CreateFont(-11, 0, 0, 0, FW_NORMAL, 0, 0, 0, DEFAULT_CHARSET,
                           OUT_DEFAULT_PRECIS, CLIP_CHARACTER_PRECIS,
                           PROOF_QUALITY, FF_DONTCARE, "MS Sans Serif");
    ci->hfontOrg   = ci->hfont;
    ci->radius     = GetSystemMetrics(SM_CYCAPTION) / 2;
    ci->border     = 1;
    ci->tailaspect = -150;
    ci->tailjoin   = 25;
    ci->joinwidth  = GetSystemMetrics(SM_CYCAPTION) / 2;
    ci->tailheight = 2 * ci->joinwidth;
    ci->BackColor  = GetNearestColor(hdc, RGB(255,255,192));
    ci->TextColor  = GetNearestColor(hdc, RGB(0,0,0));
    #if defined HTMLSUPPORT
      ci->DrawText = DrawHTML;
    #else
      ci->DrawText = DrawText;
    #endif
    ReleaseDC(hwnd, hdc);
    SetTimer(hwnd, TIMER_ID, 100, NULL);
    SetProp(hwnd, ATOM_NAME, (HANDLE)ci);
  } /* if */

  UNUSED_PARAM(lpCreateStruct);
  return TRUE;
}

static HBRUSH Cls_OnCtlColor(HWND hwnd, HDC hdc, HWND hwndChild, int type)
{
  CalloutInfo *ci = (CalloutInfo*)GetProp(hwnd, ATOM_NAME);
  SetBkMode(hdc, TRANSPARENT);
  SetTextColor(hdc, ci->TextColor);
  if (ci->hbrBackgr)
    DeleteObject(ci->hbrBackgr);
  ci->hbrBackgr = CreateSolidBrush(ci->BackColor);

  UNUSED_PARAM(hwndChild);
  UNUSED_PARAM(type);
  return ci->hbrBackgr;
}

static void Cls_OnDestroy(HWND hwnd)
{
  CalloutInfo *ci = (CalloutInfo*)GetProp(hwnd, ATOM_NAME);
  if (ci) {
    if (ci->hfontOrg)
      DeleteObject(ci->hfontOrg);
    if (ci->hbrBackgr)
      DeleteObject(ci->hbrBackgr);
    free(ci);
  } /* if */
  RemoveProp(hwnd, ATOM_NAME);
  KillTimer(hwnd, TIMER_ID);
}

static void Cls_OnPaint(HWND hwnd)
{
  CalloutInfo *ci = (CalloutInfo*)GetProp(hwnd, ATOM_NAME);
  PAINTSTRUCT ps;
  BeginPaint(hwnd, &ps);
  if (ci) {
    RECT rect;
    HBRUSH hbrBorder;
    HRGN hrgn;
    int textlength;

    GetClientRect(hwnd, &rect);
    hrgn=CreateRectRgn(0,0,1,1);
    GetWindowRgn(hwnd,hrgn);
    if (ci->hbrBackgr)
      DeleteObject(ci->hbrBackgr);
    ci->hbrBackgr = CreateSolidBrush(ci->BackColor);
    hbrBorder = CreateSolidBrush(ci->TextColor);
    FillRgn(ps.hdc, hrgn, ci->hbrBackgr);
    FrameRgn(ps.hdc, hrgn, hbrBorder, ci->border, ci->border);
    DeleteObject(hbrBorder);
    DeleteObject(hrgn);

    textlength = GetWindowTextLength(hwnd);
    if (textlength > 0) {
      LPSTR text = malloc(textlength + 1);
      if (text) {
        HFONT hfontOrg = SelectObject(ps.hdc, ci->hfont);
        GetWindowText(hwnd, text, textlength + 1);
        SetBkMode(ps.hdc, TRANSPARENT);
        SetTextColor(ps.hdc, ci->TextColor);
        ci->DrawText(ps.hdc, text, textlength, &ci->rcText,
                       DT_NOPREFIX | DT_WORDBREAK);
        SelectObject(ps.hdc, hfontOrg);
        free(text);
      } /* if */
    } /* if */

    if (ci->hicon) {
      int width = GetSystemMetrics(SM_CXSMICON);
      int height = GetSystemMetrics(SM_CYSMICON);
      DrawIconEx(ps.hdc,
                 ci->rcText.left - width - MARGIN_HOR/2, ci->rcText.top,
                 ci->hicon, width, height, 0, NULL, DI_NORMAL);
    } /* if */
  } /* if */
  EndPaint(hwnd, &ps);
}

static void Cls_OnLButtonUp(HWND hwnd, int x, int y, UINT keyFlags)
{
  /* forward to the parent or owner window */
  HWND hwndParent = GetParent(hwnd);
  NMHDR nmhdr;
  nmhdr.hwndFrom = hwnd;
  nmhdr.idFrom = 0;

  if (IsWindow(hwndParent)) {
    nmhdr.code = NM_CLICK;
    if (SendMessage(hwndParent, WM_NOTIFY, 0, (LPARAM)(LPNMHDR)&nmhdr))
      return;           /* parent handled the message, do nothing */
  } /* if */

  /* no parent, or parent did not handle the message */
  ShowWindow(hwnd, SW_HIDE);
  if (IsWindow(hwndParent)) {
    nmhdr.code = CN_POP;
    SendMessage(hwndParent, WM_NOTIFY, 0, (LPARAM)(LPNMHDR)&nmhdr);
  } /* if */

  UNUSED_PARAM(x);
  UNUSED_PARAM(y);
  UNUSED_PARAM(keyFlags);
}

static void Cls_OnRButtonUp(HWND hwnd, int x, int y, UINT keyFlags)
{
  /* forward to the parent or owner window */
  HWND hwndParent = GetParent(hwnd);
  if (IsWindow(hwndParent)) {
    NMHDR nmhdr;
    nmhdr.hwndFrom = hwnd;
    nmhdr.idFrom = 0;
    nmhdr.code = NM_RCLICK;
    SendMessage(hwndParent, WM_NOTIFY, 0, (LPARAM)(LPNMHDR)&nmhdr);
  } /* if */

  UNUSED_PARAM(x);
  UNUSED_PARAM(y);
  UNUSED_PARAM(keyFlags);
}

static void Cls_OnSetText(HWND hwnd, LPCTSTR lpszText)
{
    FORWARD_WM_SETTEXT(hwnd, lpszText, DefWindowProc);
    CalcShowCallout(hwnd, TRUE, FALSE);
}

static void Cls_OnShowWindow(HWND hwnd, BOOL fShow, UINT status)
{
  CalloutInfo *ci = (CalloutInfo*)GetProp(hwnd, ATOM_NAME);
  if (fShow && !IsWindowVisible(hwnd) && ci) {
    ci->timestamp = GetTickCount();
    CalcShowCallout(hwnd, FALSE, TRUE);
  } /* if */
  UNUSED_PARAM(status);
}

static void Cls_OnTimer(HWND hwnd, UINT id)
{
  CalloutInfo *ci = (CalloutInfo*)GetProp(hwnd, ATOM_NAME);
  if (IsWindowVisible(hwnd) && ci && ci->timeout != 0) {
    DWORD curtime = GetTickCount();
    if (curtime >= ci->timestamp + ci->timeout) {
      HWND hwndParent = GetParent(hwnd);
      ShowWindow(hwnd, SW_HIDE);
      if (IsWindow(hwndParent)) {
        NMHDR nmhdr;
        nmhdr.hwndFrom = hwnd;
        nmhdr.idFrom = 0;
        nmhdr.code = CN_POP;
        SendMessage(hwndParent, WM_NOTIFY, 0, (LPARAM)(LPNMHDR)&nmhdr);
      } /* if */
    } /* if */
  } /* if */
  UNUSED_PARAM(id);
}

LRESULT CALLBACK CalloutProc(HWND hwnd, UINT message,
                             WPARAM wParam, LPARAM lParam )
{
  CalloutInfo *ci;
  switch (message) {
    HANDLE_MSG(hwnd, WM_CREATE,      Cls_OnCreate);
    HANDLE_MSG(hwnd, WM_CTLCOLORBTN, Cls_OnCtlColor);
    HANDLE_MSG(hwnd, WM_DESTROY,     Cls_OnDestroy);
    HANDLE_MSG(hwnd, WM_LBUTTONUP,   Cls_OnLButtonUp);
    HANDLE_MSG(hwnd, WM_PAINT,       Cls_OnPaint);
    HANDLE_MSG(hwnd, WM_RBUTTONUP,   Cls_OnRButtonUp);
    HANDLE_MSG(hwnd, WM_SETTEXT,     Cls_OnSetText);
    HANDLE_MSG(hwnd, WM_SHOWWINDOW,  Cls_OnShowWindow);
    HANDLE_MSG(hwnd, WM_TIMER,       Cls_OnTimer);

    case WM_COMMAND:
    case WM_NOTIFY:
      {
        HWND hwndParent = GetParent(hwnd);
        if (IsWindow(hwndParent))
          return SendMessage(hwndParent, message, wParam, lParam);
      }
      return 0;

    case WM_GETFONT:
      return (LRESULT)GETFIELD(hfont);
    case WM_SETFONT:
      SETFIELD(hfont, (HFONT)wParam, LOWORD(lParam));
      return 0L;
    case WM_GETICON:
      return (wParam == ICON_SMALL) ? (LRESULT)GETFIELD(hicon) : 0;
    case WM_SETICON:
      if (wParam == ICON_SMALL) {
        HICON prevIcon = GETFIELD(hicon);
        SETFIELD(hicon, (HICON)lParam, TRUE);
        return (LRESULT)prevIcon;
      } /* if */
      return 0;

    case CM_GETANCHOR:
      if ((ci=(CalloutInfo*)GetProp(hwnd,ATOM_NAME))!=NULL) {
        POINT pt = ci->ptAnchor;
        HWND hwndParent = GetParent(hwnd);
        if (IsWindow(hwndParent))
          ScreenToClient(hwndParent, &pt);  /* translate coordinates */
        return MAKELONG(pt.x, pt.y);
      } /* if */
      return 0;
    case CM_SETANCHOR:
      if ((ci=(CalloutInfo*)GetProp(hwnd,ATOM_NAME))!=NULL) {
        HWND hwndParent = GetParent(hwnd);
        POINT pt;
        pt.x = LOWORD(lParam);
        pt.y = HIWORD(lParam);
        if (IsWindow(hwndParent))
          ClientToScreen(hwndParent, &pt);  /* translate coordinates */
        ci->ptAnchor = pt;
      } /* if */
      return 0;
    case CM_GETRECT:
      CalcShowCallout(hwnd, FALSE, TRUE);
      return ((ci=(CalloutInfo*)GetProp(hwnd,ATOM_NAME))!=NULL)
               ? (LRESULT)&ci->rcText : 0;

    HANDLEPROP(MINWIDTH,    minwidth,    (int)wParam,      LOWORD(lParam));
    HANDLEPROP(EXTRAHEIGHT, extraheight, (int)wParam,      LOWORD(lParam));
    HANDLEPROP(RADIUS,      radius,      (int)wParam,      LOWORD(lParam));
    HANDLEPROP(BORDER,      border,      (int)wParam,      LOWORD(lParam));
    HANDLEPROP(TAILSTYLE,   tailstyle,   (int)wParam,      LOWORD(lParam));
    HANDLEPROP(TAILSLANT,   tailaspect,  (int)wParam,      LOWORD(lParam));
    HANDLEPROP(TAILHEIGHT,  tailheight,  (int)wParam,      LOWORD(lParam));
    HANDLEPROP(TAILOFFSET,  tailoffset,  (int)wParam,      LOWORD(lParam));
    HANDLEPROP(TAILJOIN,    tailjoin,    (int)wParam,      LOWORD(lParam));
    HANDLEPROP(TAILWIDTH,   joinwidth,   (int)wParam/2,    LOWORD(lParam));
    HANDLEPROP(TEXTCOLOR,   TextColor,   (COLORREF)wParam, LOWORD(lParam));
    HANDLEPROP(BACKCOLOR,   BackColor,   (COLORREF)wParam, LOWORD(lParam));
    HANDLEPROP(VERTALIGN,   alignbelow,  (int)wParam,      LOWORD(lParam));
    HANDLEPROP(DRAWTEXTPROC,DrawText,    (DRAWTEXT)wParam, LOWORD(lParam));
    HANDLEPROP(TIMEOUT,     timeout,     (DWORD)lParam,    FALSE);

  default:
    return DefWindowProc(hwnd, message, wParam, lParam);
  } /* switch */
}

BOOL CalloutInit(HINSTANCE hInstance)
{
  WNDCLASS wc;

  /* Fill in the window class structure that describes the "tool tip" */
  memset(&wc, 0, sizeof wc);
  wc.style = CS_GLOBALCLASS;
  wc.lpfnWndProc = CalloutProc;
  wc.hInstance = hInstance;
  wc.hCursor = LoadCursor(NULL, IDC_ARROW);
  wc.lpszClassName = CALLOUT_CLASS;
  return RegisterClass(&wc);
}

void CalloutCleanup(HINSTANCE hInstance)
{
  UnregisterClass(CALLOUT_CLASS, hInstance);
}

#if defined CALLOUT_DLL

#if defined(__WATCOMC__)
  #define DllMain(h,dw,lp)      LibMain(h,dw,lp)
#elif defined(__BORLANDC__)
  #define DllMain(h,dw,lp)      DllEntryPoint(h,dw,lp)
#endif

BOOL WINAPI DllMain(HINSTANCE hInstance, DWORD dwReason, LPVOID lpRes)
{
  switch (dwReason) {
  case DLL_PROCESS_ATTACH:
    CalloutInit(hInstance);
    break;

  case DLL_PROCESS_DETACH:
    break;
  } /* switch */

  UNUSED_PARAM(lpRes);
  return TRUE;
}

#endif /* CALLOUT_DLL */
